/*

Copyright (C) 2003 Frank Condello.

Please see LICENSE.txt included with this distribution for usage restrictions.

*/

/* ZZipUtilities.cpp: Main plugin functions and entry point */

#include "RBCommon.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if TARGET_OS_MAC
# if defined(__MWERKS__)
#  include <FSp_fopen.h>
# elif defined(__GNUC__)
#  include <sys/param.h>
# endif
#else
 typedef unsigned char Byte;// Workaround for a problem with zlib headers in Win32 build
#endif

#include <zzip/zzip.h>

/* Prototypes */
static REALstring ZZipExtractFileData (REALobject instance, REALstring path);
static REALstring ZZipLastDirError (REALobject instance);
static REALobject ZZipOpenAsZZipArchive (REALobject instance);
static void ZZipArchiveConstructor (REALobject instance);
static void ZZipArchiveDestructor (REALobject instance);
static FILE *_getANSIFile(REALfolderItem theFile, const char *fileMode);
static int _getFileLength (ZZIP_FILE *f);


/* ZZipArchive class methods */
static REALmethodDefinition ZZipArchiveMethods[] = {
	{ (REALproc)ZZipExtractFileData, REALnoImplementation, "ExtractFileData(path As String) As String" },
	{ (REALproc)ZZipLastDirError, REALnoImplementation, "LastError() As String" }
	
};


/* Folderitem extention */
static REALmethodDefinition ZZipFolderitemMethods[] = {
	{ (REALproc)ZZipOpenAsZZipArchive, REALnoImplementation, "OpenAsZZipArchive(Extends f As Folderitem) As ZZipArchive" }
};


/* ZZipArchive class internal storage */
struct ZZipArchiveClassData
{
	ZZIP_DIR	*zipFile;
};


/* ZZipArchive class */
static REALclassDefinition ZZipArchiveClass = {
	kCurrentREALControlVersion,
	"ZZipArchive",
	"",									// superName
	sizeof(ZZipArchiveClassData),		// data size
	0,									// reserved
	(REALproc)ZZipArchiveConstructor,	// constructor
	(REALproc)ZZipArchiveDestructor,	// destructor
	NULL,								// properties
	0,									// property count
	ZZipArchiveMethods,					// methods
	SizeOfMethods(ZZipArchiveMethods),	// method count
	NULL,								// events
	0,									// event count
	NULL,								// event instances
	0,									// event instance count
	NULL								// interface(s)
};


#pragma mark -


/* ---------------------------------------------------------------------------
	ZZipOpenAsZZipArchive
	
	Folderitem extension - Attempts to open theFile as a zip archive, and
	returns a ZZipArchive object if successful.
---------------------------------------------------------------------------- */
static REALobject ZZipOpenAsZZipArchive (REALobject instance)
{
	FILE			*ansiFile;
	ZZIP_DIR 		*zipDir;
	zzip_error_t	err;
	REALobject		zzarch = NULL;
	REALfolderItem	theFile = (REALfolderItem)instance;

	ansiFile = _getANSIFile(theFile, "rb");
	
	if (ansiFile)
	{
		zipDir = zzip_dir_fdopen(fileno(ansiFile), &err);

		if (zipDir)
		{
			zzarch = REALnewInstance("ZZipArchive");
			if (zzarch)
			{
				ClassData(ZZipArchiveClass, zzarch, ZZipArchiveClassData, cdata);
				cdata->zipFile = zipDir;
			}
			else
				zzip_dir_close(zipDir);
		}
		else
			fclose(ansiFile);
	}
	
	return zzarch;
}


/* ---------------------------------------------------------------------------
	ZZipExtractFileData
	
	Returns uncompressed file data as a string.
---------------------------------------------------------------------------- */
static REALstring ZZipExtractFileData (REALobject instance, REALstring path)
{
	ZZIP_FILE	*f;
	int			len;
	char		*buf = NULL;
	REALstring	fdata = NULL;
	
	REALLockObject(instance);
	
	ClassData (ZZipArchiveClass, instance, ZZipArchiveClassData, cdata);

	if (cdata->zipFile)
	{
		f = zzip_file_open(cdata->zipFile, REALCString(path), ZZIP_CASELESS);

		if (f)
		{
			len = _getFileLength(f);
			if (len)
				buf = (char *)malloc(len);
			
			if (!buf)
			{
				zzip_file_close(f);
				return NULL;
			}

			zzip_file_read(f, buf, len);
			zzip_file_close(f);
			
			fdata = REALBuildString(buf, len);
			
			free(buf);
		}
	}
	
	REALUnlockObject(instance);
	
	return fdata;
}


/* ---------------------------------------------------------------------------
	ZZipLastDirError
	
	Get the last error string for a ZZIP_DIR
	FIXME! zzip_strerror_of should be called in ZZipExtractFileData
	and be properly filled with the archive name and/or file path
	depending on the error - This method should then return a premade,
	cached, RB String.
---------------------------------------------------------------------------- */
static REALstring ZZipLastDirError (REALobject instance)
{
	REALstring	err = NULL;
	char		*strerr;

	REALLockObject(instance);

	ClassData (ZZipArchiveClass, instance, ZZipArchiveClassData, cdata);

	if (cdata->zipFile)
	{
		strerr = (char *)zzip_strerror_of(cdata->zipFile);
		err = REALBuildString(strerr, strlen(strerr));
	}
	
	REALUnlockObject(instance);

	return err;
}


/* ---------------------------------------------------------------------------
	ZZipArchiveConstructor
---------------------------------------------------------------------------- */
static void ZZipArchiveConstructor (REALobject instance)
{
	ClassData (ZZipArchiveClass, instance, ZZipArchiveClassData, cdata);
	cdata->zipFile = NULL;
}


/* ---------------------------------------------------------------------------
	ZZipArchiveDestructor
---------------------------------------------------------------------------- */
static void ZZipArchiveDestructor (REALobject instance)
{
	ClassData (ZZipArchiveClass, instance, ZZipArchiveClassData, cdata);
	if (cdata->zipFile) zzip_dir_close(cdata->zipFile);
}


#pragma mark -


/* ---------------------------------------------------------------------------
	_getANSIFile
	
	Get a FILE handle from a folderitem.
	NOTE: This is read only on Mac OS (that's all we need for this plug).
---------------------------------------------------------------------------- */
#define MAXFILEPATH 2048
static FILE *_getANSIFile(REALfolderItem theFile, const char *fileMode)
{
	FILE	*ansiFile = NULL;
	
	if (!theFile) return NULL;
	
#if TARGET_OS_MAC
	
	FSSpec			theFSSpec;
	FSRef			theFSRef;
	char			*thePath = NULL;

	// Grab the FSSpec
	if (!REALFSSpecFromFolderItem(&theFSSpec, theFile))
		return(NULL);

	// And open the file...
	if ( FSpMakeFSRef(&theFSSpec, &theFSRef) == noErr )
	{
		thePath = (char *)malloc(MAXFILEPATH);
		if (thePath)
		{
#if defined(__GNUC__)
			if ( FSRefMakePath(&theFSRef, (UInt8 *)thePath, MAXFILEPATH) == noErr )
				ansiFile = fopen((const char*)thePath, fileMode);
#elif defined(__MWERKS__)
			ansiFile = FSRef_fopen(&theFSRef, fileMode);
#endif
			free(thePath);
		}
	}

#else
	
	// Use the absolute path on Win32/Linux
	REALstring	path = NULL;

	path = REALConvertString( REALpathFromFolderItem(theFile), REALGetFontEncoding("System") );
	if (path) ansiFile = fopen(path->CString(), fileMode);

#endif

	return ansiFile;
}


/* ---------------------------------------------------------------------------
	_getFileLength
	
	Return the uncompressed file length. Don't touch cursor position.
---------------------------------------------------------------------------- */
static int _getFileLength (ZZIP_FILE *f)
{
	int	pos;
	int	end;

	pos = zzip_tell (f);
	zzip_seek (f, 0, SEEK_END);
	end = zzip_tell (f);
	zzip_seek (f, pos, SEEK_SET);

	return end;
}


#pragma mark -


/* ---------------------------------------------------------------------------
	PluginEntry
---------------------------------------------------------------------------- */
void PluginEntry(void)
{
	UInt32		n;
		
	REALRegisterClass(&ZZipArchiveClass);

	// Register the global methods
	for (n = 0; n < ( SizeOfMethods(ZZipFolderitemMethods) ); n++)
		REALRegisterMethod(&ZZipFolderitemMethods[n]);
}